/* ir_remote_control

This scetch records and reprduces IR signals from IR remote controls.
The IR data is stored in NVS memory and therefore retaiend beyond power off.
This version offers 3 channels. 

Pin assignments:
  bt_learn        input pullup    GPIO16          learn button, push for at least 3 seconds
  bt_play         input pullup    GPIO17, 5, 18   buttons to send signal 
  ir receiver     input           GPIO34          IR receiver chip, e.g.TSOP1138 
  ir transmitter  output          GPIO33          IR transmitter diode, connected via a transistor
  led             output          GPIO27, 26, 25  control leds
  
SLW 10-Sep-2019
*/

#include <Arduino.h>
#include <Preferences.h>

// pin assignments
const int bt_learn = 16;
const int bt_play[] = {17, 5, 18};
const int ir_receiver = 34, ir_transmitter = 33;
const int led[] = {27, 26, 25};

// global constants 
const int channel_cnt = 3;                  // this version supports 3 channels (maximum is 10)
const uint16_t ir_data_size = 250;          // size of ir data sequence
const uint16_t ir_pwm_frequency = 38000;    // ir signal sends at 38kHz
const uint8_t ir_pwm_channel = 0;           // first pwm channel for ir signal
const uint8_t led_pwm_channel = 2;          // second pwm channel for LEDs
const uint16_t led_pwm_frequency = 8;       // frequency for flashing LEDs
const uint32_t ir_read_timeout = 250000;    // max duration is 0.2 sec 

// global variables
Preferences prefs;                            // access to non volatile storage (NVS)
uint32_t ir_data[channel_cnt][ir_data_size];  // IR data set, 3 rows of up to 250 data items each
int ir_data_len[channel_cnt] = {0, 0, 0};
bool learn_bt_pressed = false;

/* startup_msg ---------------------------------------------------------------------------
Flashes the LEDs at startup, just to show that the system is active.
*/
void startup_msg(void) {
  for (int i = 0; i < channel_cnt; ++i) {
    digitalWrite(led[i], LOW);
    delay(500);
    digitalWrite(led[i], HIGH);
  }
}

/* read_ir_data ---------------------------------------------------------------------------
Reads a sequence of ir data via the IR receiver. Captures the start- and stop time points
of each burst and stores them to the array ir_data[]. Data reading continues until either 
the data storage is exceeded or reading timeout is reached. Flashes the corresponding led
during the reading process. 
Input:    channel number
Output:   array ir_data -> ir data points
          int *ir_data_len -> number of data points stored in the array 
*/
void read_ir_data(int channel, uint32_t ir_data[], int *ir_data_len) {
  long start_time;
  uint32_t time_stamp;
  bool edge = true;            // false for falling edge, true for rising edge

  ledcAttachPin(led[channel], led_pwm_channel);                     // flash LED
  ledcWrite(led_pwm_channel, 128);
  *ir_data_len = 0;                                                    // reset array index
  while (digitalRead(ir_receiver) == HIGH);                         // wait for falling edge
  start_time = esp_timer_get_time();                                // capture start time
  do {
    time_stamp = (uint32_t) (esp_timer_get_time() - start_time);    // refresh time stamp 
    if (edge) {                                                     
      if (digitalRead(ir_receiver) == HIGH) {                       // waiting for rising edge
        ir_data[*ir_data_len] = time_stamp;                            // capture time stamp
        ++*ir_data_len;
        edge = false;                                               // switch to falling edge
      };  
    } else {                                                       
      if (digitalRead(ir_receiver) == LOW) {                        // waiting for falling edge
        ir_data[*ir_data_len] = time_stamp;                            // capture time stamp 
        ++*ir_data_len;
        edge = true;                                                // switch to rising edge
      };  
    };
  } while ((time_stamp < ir_read_timeout) && (*ir_data_len < ir_data_size));
  ledcDetachPin(led[channel]);                                      // clear flashing LED
  digitalWrite(led[channel], HIGH);                                 // switch off LED
}

/* play_ir_data ---------------------------------------------------------------------------
Plays an IR data sequence
Input:  array ir_data -> holds the start- and stop time points for the bursts
        int ir_data_len -> number of data points in the array
*/
void play_ir_data(uint32_t ir_data[], uint16_t ir_data_len) {
  long start_time;
  uint32_t time_stamp;
  bool edge = true;                     // true for active, false for pause
  int i = 0;

  if (ir_data_len == 0) return;         // if data set is empty, then nothing to do

  start_time = esp_timer_get_time();    // capture start time
  ledcWrite(ir_pwm_channel, 128);       // activate IR transmitter
  do {
    time_stamp = esp_timer_get_time() - start_time;
    if (time_stamp >= ir_data[i]) {
      if (edge) {
         ledcWrite(ir_pwm_channel, 0);    // stop IR transmitter
         edge = false;
      } else {
         ledcWrite(ir_pwm_channel, 128);  // activate IR transmitter
         edge = true;
      };
      ++i;
    }
  } while (i < ir_data_len);
  ledcWrite(ir_pwm_channel, 0);           // ensure that IR transmitter is stopped 
}

/* print_ir_data ---------------------------------------------------------------------------
Prints ir data to the serial monitor, for control purpose only
*/
void print_ir_data(int channel, uint32_t ir_data[], uint16_t ir_data_len) {
  int line_cnt = 0;

  Serial.println();
  Serial.print("Channel " + String(channel) + "  IR data (");
  Serial.print(ir_data_len);
  Serial.println("):");
  for (int i = 0; i < ir_data_len; ++i) {
    Serial.print(ir_data[i]);
    if (i < ir_data_len - 1) Serial.print(", ");
    ++line_cnt;
    if (line_cnt > 10) {
      Serial.println();
      line_cnt = 0;
    }
  }
  Serial.println();
}

/* -------------------------------------------------------------------------------------*/
void learn_sequence(void) {
  char ir_len_str[10] = "ir_len_0", ir_data_str[10] = "ir_data_0";
  for (int i = 0; i < channel_cnt; ++i) {
    read_ir_data(i, ir_data[i], &ir_data_len[i]);
    print_ir_data(i, ir_data[i], ir_data_len[i]);
    ir_len_str[7] = '0' + i;
    ir_data_str[8] = '0' + i;
    prefs.putUShort(ir_len_str, ir_data_len[i]);
    prefs.putBytes(ir_data_str, ir_data[i], ir_data_len[i] * sizeof(uint32_t));
  };
}

/* -------------------------------------------------------------------------------------*/
void setup() {
  char ir_len_str[10] = "ir_len_0", ir_data_str[10] = "ir_data_0";

  Serial.begin(115200);

  // read IR data from NVS
  prefs.begin("ir_nvs", false);
  Serial.println();
  Serial.println("Reading NVS ir_data ");
  for (int i = 0; i < channel_cnt; ++i) {
    ir_len_str[7] = '0' + i;
    ir_data_str[8] = '0' + i;
    ir_data_len[i] = prefs.getUShort(ir_len_str);
    Serial.print("Channel " + String(i) + ": ");
    Serial.println(String(ir_data_len[i]) + " Entries");
    if (ir_data_len[i] > 0) 
      prefs.getBytes(ir_data_str, ir_data[i], ir_data_len[i] * sizeof(uint32_t));
  }       
  Serial.println();

  // set pin modes
  for (int i = 0; i < channel_cnt; ++i) {
    pinMode(bt_play[i], INPUT_PULLUP);        // define ports for buttons
    pinMode(led[i], OUTPUT);                  // define ports for LEDs
    digitalWrite(led[i], HIGH);               // switch off the LEDs
  };
  pinMode(bt_learn, INPUT_PULLUP);
  pinMode(ir_receiver, INPUT);
 
  // setup PWM for IR transmitter
  ledcSetup(ir_pwm_channel, ir_pwm_frequency, 8);    // pwm runs with 8 bit resolution
  ledcAttachPin(ir_transmitter, ir_pwm_channel);
  ledcWrite(ir_pwm_channel, 0);                      // switch off the IR transmitter, for now

  // setup PWM for flashing LEDs
  ledcSetup(led_pwm_channel, led_pwm_frequency, 8);

  // show welcome message
  startup_msg();
}

/* -------------------------------------------------------------------------------------*/
void loop() {
  uint8_t learn_pressed_cnt = 0;

  // check whether the learn button is pressed for at least 3 seconds
  while (digitalRead(bt_learn) == LOW) {
    delay(200);
    ++learn_pressed_cnt;
    if (learn_pressed_cnt >= 15) {
      learn_sequence();
      learn_pressed_cnt = 0;
    }
  }

  // check the play buttons and play correspoding sequence
  for (int i = 0; i < channel_cnt; ++i) {  
    if (digitalRead(bt_play[i]) == LOW) {
      digitalWrite(led[i], LOW);
      play_ir_data(ir_data[i], ir_data_len[i]);
      digitalWrite(led[i], HIGH);
      delay(250);
    }
  }
}
